import inspect
import os
from argparse import ArgumentParser
from pathlib import Path
from typing import Optional, List

from rick_db.conn import Connection

from pokie.constants import DI_DB, DI_APP
from pokie.core import CliCommand
from rick_db.util import MigrationRecord


class DbCliCommand(CliCommand):
    # migrations folder
    folder = Path('sql')

    # common error messages
    error_nodb = "error: no database connection found in the application"
    error_noinit = "error: migration manager not installed; run 'dbinit' command first"

    def get_db(self) -> Optional[Connection]:
        di = self.get_di()
        if not di.has(DI_DB):
            return None
        return di.get(DI_DB)

    def load_migrations(self, module_name: str, path: Path) -> List[tuple]:
        """
        Scan path for sql files and loads contents into a list
        :param module_name: module name to use
        :param path: path to scan
        :return: list of (MigrationRecord, content)
        """
        mig_dict = {}
        for entry in sorted(path.glob('*.sql')):
            if entry.is_file():
                with open(entry, encoding="utf-8") as f:
                    mig_dict[entry.name] = f.read()

        result = []
        for name, contents in mig_dict.items():
            record = MigrationRecord(name="{}/{}".format(module_name, name))
            result.append((record, contents))
        return result


class DbInitCmd(DbCliCommand):
    description = "initialize database migrations"

    def run(self, args) -> bool:
        db = self.get_db()
        if not db:
            self.tty.error(self.error_nodb)
            return False

        mgr = db.migration_manager()
        if not mgr.has_manager():
            self.tty.write("installing migration manager...", False)
            result = mgr.install_manager()
            if result.success:
                self.tty.write(self.tty.colorizer.green("success"))
                return True
            else:
                self.tty.write(self.tty.colorizer.green("error: " + result.error))
                return False

        self.tty.write("migration manager already installed")
        return True


class DbCheckCmd(DbCliCommand):
    description = "show existing database migrations status"

    def run(self, args) -> bool:
        db = self.get_db()
        if not db:
            self.tty.error(self.error_nodb)
            return False

        mgr = db.migration_manager()
        if not mgr.has_manager():
            self.tty.error(self.error_noinit)
            return False

        for name, module in self.get_di().get(DI_APP).modules.items():
            self.tty.write("Checking migrations for module {}:".format(name))
            path = Path(os.path.dirname(inspect.getfile(module.__class__))) / self.folder
            if path.exists() and path.is_dir():
                try:
                    for record in self.load_migrations(name, path):
                        mig, content = record
                        self.tty.write("\t{name}... ".format(name=mig.name), False)

                        # check if migration is duplicated
                        record = mgr.fetch_by_name(mig.name)
                        if record is not None:
                            self.tty.write(self.tty.colorizer.white("already applied", attr='bold'))

                        # check if migration is obviously empty
                        elif content.strip() == "":
                            self.tty.write(self.tty.colorizer.yellow("empty migration", attr='bold'))
                        else:
                            self.tty.write(self.tty.colorizer.green("new migration", attr='bold'))

                except Exception as e:
                    self.tty.error("Error : " + str(e))
                    return False
        return True

class DbUpdateCmd(DbCliCommand):
    description = "apply pending migrations"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument('--dry', help='Dry run - no database changes are performed', action='store_true',
                            default=False)

    def run(self, args) -> bool:
        db = self.get_db()
        if not db:
            self.tty.error(self.error_nodb)
            return False

        mgr = db.migration_manager()
        if not mgr.has_manager():
            self.tty.error(self.error_noinit)
            return False

        for module_name, module in self.get_di().get(DI_APP).modules.items():
            self.tty.write("Checking migrations for module {}:".format(module_name))
            path = Path(os.path.dirname(inspect.getfile(module.__class__))) / self.folder
            if path.exists() and path.is_dir():
                try:
                    for record in self.load_migrations(module_name, path):
                        mig, content = record
                        self.tty.write("\t{name}... ".format(name=mig.name), False)

                        # check if migration is duplicated
                        record = mgr.fetch_by_name(mig.name)
                        if record is not None:
                            self.tty.write(self.tty.colorizer.white("already applied", attr='bold'))

                        # check if migration is obviously empty
                        elif content.strip() == "":
                            self.tty.write(self.tty.colorizer.yellow("empty migration", attr='bold'))
                        else:
                            # apply migration
                            if not args.dry:
                                # try to execute migration and register on the migration manager
                                result = mgr.execute(mig, content)
                                if result.success:
                                    self.tty.write(self.tty.colorizer.green("success", attr='bold'))
                                else:
                                    # in case of error, abort
                                    self.tty.write("\n")
                                    self.tty.error("Error: " + result.error)
                                    return False
                            else:
                                # dry run, just assume everyting is fine
                                self.tty.write(self.tty.colorizer.green("success", attr='bold'))

                except Exception as e:
                    self.tty.write("\n")
                    self.tty.error("Error : " + str(e))
                    return False
        return True



from .base import ListCmd, HelpCmd, RunServerCmd, VersionCmd
from .db import DbInitCmd, DbCheckCmd, DbUpdateCmd
from .job import JobRunCmd, JobListCmd
from .codegen import GenDtoCmd, GenRequestRecordCmd


import json
import logging

from flask.views import MethodView
from flask import current_app, jsonify
from flask_login import current_user
from rick.serializer.json import ExtendedJsonEncoder
from rick.form import RequestRecord

from .response import JsonRequestError, JsonStatus
from pokie.constants import HTTP_OK, HTTP_BADREQ, HTTP_INTERNAL_ERROR, HTTP_NOAUTH, HTTP_FORBIDDEN, DI_SERVICE_MANAGER


class PokieView(MethodView):

    def __init__(self):
        super().__init__()
        self.di = current_app.di

    def json(self, data, code=HTTP_OK):
        """
        Adaptation of flask's jsonify using Rick's ExtendedJsonEncoder
        :param data:
        :param code:
        :return:
        """
        indent = None
        separators = (",", ":")

        if current_app.config["JSONIFY_PRETTYPRINT_REGULAR"] or current_app.debug:
            indent = 2
            separators = (", ", ": ")

        data = json.dumps(data, indent=indent, separators=separators, cls=ExtendedJsonEncoder)
        return current_app.response_class(data, status=code, mimetype=current_app.config["JSONIFY_MIMETYPE"])

    def error(self, message=None, code=HTTP_BADREQ):
        data = JsonStatus(success=False, message=message if message else "operation failed")
        return self.json(data, code)

    def request_error(self, req: RequestRecord, code=HTTP_BADREQ):
        return self.json(JsonRequestError(success=False, formError=req.get_errors()), code)

    def success_message(self, message=''):
        return self.json(JsonStatus(success=True, message=message))

    def success(self, data=None, code=HTTP_OK):
        if data is None:
            return self.json(JsonStatus(success=True, message=''), code=code)
        return self.json(data, code)

    def empty_body(self):
        return self.error('empty body', code=HTTP_BADREQ)

    def not_found(self):
        return self.error("record not found", code=HTTP_BADREQ)

    def forbidden(self):
        return self.error("access denied", code=HTTP_FORBIDDEN)

    def denied(self):
        return self.error("access denied", code=HTTP_NOAUTH)

    def get_service(self, service_name):
        return self.di.get(DI_SERVICE_MANAGER).get(service_name)


class PokieAuthView(PokieView):
    # list of acls to check for current user
    # if list is empty, no acl control is used
    acl = []
    user = None

    def __init__(self):
        super().__init__()
        self.user = current_user

    def dispatch_request(self, *args, **kwargs):

        if not current_user.is_authenticated:
            return self.denied()

        for acl in self.acl:
            if not self.user.can_access(acl):
                return self.forbidden()

        try:
            return super().dispatch_request(*args, **kwargs)

        except Exception as e:
            logging.exception(e)
            return self.error("there was an internal error processing the request", code=HTTP_INTERNAL_ERROR)



import os
import sys
from argparse import ArgumentParser

from rick.resource.console import AnsiColor
from rick.util.loader import load_class

import pokie
from pokie.constants import DI_APP, DI_FLASK
from pokie.core import CliCommand
from pokie.util.cli_args import ArgParser


class BaseCommand(CliCommand):

    def get_modules(self) -> dict:
        return self.get_di().get(DI_APP).modules

    def get_cmd_map(self) -> dict:
        result = {}
        for _, module in self.get_modules().items():
            for cmd, cmd_class in module.cmd.items():
                result[cmd] = cmd_class
        return result

    def run(self, args) -> bool:
        di = self.get_di()
        color = AnsiColor()
        self.tty.write("Available commands:\n")
        for _, module in di.get(DI_APP).modules.items():
            for cmd, cmd_path in module.cmd.items():
                cls = load_class(cmd_path)
                if not cls:
                    raise RuntimeError(
                        "Error: class '{}' not found while listing available CLI commands".format(cmd_path))
                if not issubclass(cls, CliCommand):
                    raise RuntimeError("Error: class '{}' does not extend CliCommand".format(cmd_path))
                obj = cls(di)
                self.tty.write("{} \t {}".format(color.green(cmd), color.white(obj.description)))

        return True


class VersionCmd(BaseCommand):
    description = "display pokie version"

    def run(self, args) -> bool:
        self.tty.write(pokie.get_version())
        return True


class HelpCmd(BaseCommand):
    description = "display usage information for a given command"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument('command', type=str, help="Command to get usage details.")

    def run(self, args) -> bool:
        map = self.get_cmd_map()
        if args.command not in map.keys():
            self.tty.error("Error: command '{}' not found".format(args.command))
            return False

        cls = load_class(map[args.command])
        handler = cls(self.get_di())  # type: CliCommand
        self.show(args.command, handler)
        return True

    def show(self, cmd, cmd_object):
        """
        Show command detail
        :param cmd: command
        :param cmd_object: command object
        :return:
        """
        program = os.path.basename(sys.argv[0])

        self.tty.write("{}: {}\n".format(cmd, cmd_object.description))
        self.tty.write("usage: {} {} [OPTIONS...]\n".format(program, cmd))

        parser = ArgParser(add_help=False)
        cmd_object.arguments(parser)
        self.tty.write(parser.format_parameters())


class ListCmd(BaseCommand):
    description = "list available commands"

    def run(self, args) -> bool:
        color = AnsiColor()
        self.tty.write("\nusage: {} <command> [OPTIONS...]\n".format(os.path.basename(sys.argv[0])))
        self.tty.write("available commands:")
        for cmd, cmd_path in self.get_cmd_map().items():
            cls = load_class(cmd_path)

            if not cls:
                raise RuntimeError("Error: class '{}' not found while listing available CLI commands".format(cmd_path))

            if not issubclass(cls, CliCommand):
                raise RuntimeError("Error: class '{}' does not extend CliCommand".format(cmd_path))

            # show details
            self.tty.write("{} \t {}".format(color.green(cmd), color.white(cls.description)).expandtabs(32))

        return True


class RunServerCmd(CliCommand):
    description = "run Flask webserver for development purposes"

    # flask_name: args_name
    flask_args = {
        'host': 'host',
        'port': 'port',
        'debug': 'debug',
        'use_reloader': 'reload'
    }

    def arguments(self, parser: ArgumentParser):
        parser.add_argument('-h', '--host', help="The interface to bind to (default: 127.0.0.1)", required=False,
                            default='127.0.0.1')
        parser.add_argument('-p', '--port', help="The port to bind to (default: 5000)", required=False, default='5000')
        parser.add_argument('-d', '--debug', help="Enable debug mode (default: false)", required=False,
                            action='store_true',
                            default=False)
        parser.add_argument('-r', '--reload', help="Enable automatic reload (default: false)", required=False,
                            action='store_true',
                            default=False)

    def run(self, args) -> bool:
        kwargs = {}
        for a, b in self.flask_args.items():
            kwargs[a] = getattr(args, b)

        # run flask
        self.get_di().get(DI_FLASK).run(**kwargs)

        return True



from typing import Any, Optional, List

from rick.base import Di
from rick.mixin import Injectable

from pokie.constants import DI_DB
from pokie.contrib.base.dto import SettingsRecord
from pokie.contrib.base.repository.settings import SettingsRepository


class SettingsService(Injectable):

    def __init__(self, di: Di):
        super().__init__(di)
        self.db = di.get(DI_DB)

    def list(self) -> List[SettingsRecord]:
        return self.repo_config.fetch_all()

    def by_module(self, module: str) -> List[SettingsRecord]:
        return self.repo_config.fetch_by_module(module)

    def by_key(self, module: str, key: str) -> Optional[SettingsRecord]:
        return self.repo_config.fetch_by_key(module, key)

    def upsert(self, module: str, key: str, value: str):
        return self.repo_config.upsert(module, key, value)

    def delete(self, module, key):
        record = self.repo_config.fetch_by_key(module, key)
        if record is not None:
            self.repo_config.delete_pk(record.id)

    @property
    def repo_config(self) -> SettingsRepository:
        return SettingsRepository(self.db)



from argparse import ArgumentParser
import getpass
from rick.form import RequestRecord, field

from pokie.contrib.auth.constants import SVC_USER, SVC_AUTH
from pokie.constants import DI_SERVICE_MANAGER
from pokie.contrib.auth.dto import UserRecord
from pokie.contrib.auth.service import UserService, AuthService
from pokie.core import CliCommand


class UserCreateRequest(RequestRecord):
    fields = {
        'username': field(validators="required|minlen:1|maxlen:200", error="invalid username"),
        'email': field(validators="required|minlen:3|maxlen:200|email", error="invalid email address"),
        'first_name': field(validators="maxlen:100", error="invalid first name"),
        'last_name': field(validators="maxlen:100", error="invalid last name"),
    }


class UserInfoRequest(RequestRecord):
    fields = {
        'username': field(validators="required|pk:user,username", error="invalid username")
    }


class UserCommand(CliCommand):

    @property
    def svc_user(self) -> UserService:
        return self.get_di().get(DI_SERVICE_MANAGER).get(SVC_USER)


class UserCreateCmd(UserCommand):
    description = "create a new user"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument('username', type=str, help="User name to create")
        parser.add_argument('email', type=str, help="User email")
        parser.add_argument('-f', '--first_name', help="First name", required=False)
        parser.add_argument('-l', '--last_name', help="First name", required=False)
        parser.add_argument('--enabled', default=True, action='store_true',
                            help="Enable user (default True)")
        parser.add_argument('--admin', default=False, action='store_true',
                            help="Create admin (default False)")

    def run(self, args) -> bool:

        req = UserCreateRequest()
        if not req.is_valid({'username': args.username, 'email': args.email}):
            for k, v in req.get_errors().items():
                self.tty.write(self.tty.colorizer.red("Error in {}: {}".format(k, v['*'])))
            return False

        if self.svc_user.get_by_username(args.username) is not None:
            self.tty.write(self.tty.colorizer.red("Error: username already exists"))
            return False

        record = UserRecord(username=args.username, email=args.email, admin=args.admin, active=args.enabled,
                            password="", first_name=args.first_name, last_name=args.last_name)
        id = self.svc_user.add_user(record)
        self.tty.write("Created user #{} with username '{}' and email '{}'".format(str(id), args.username, args.email))

        return True


class UserInfoCmd(UserCommand):
    description = "list user details"

    fields_labels = {
        'id': 'User Id',
        'username': 'Username',
        'active': 'Active',
        'admin': 'Administrator',
        'first_name': 'First name',
        'last_name': 'Last name',
        'email': 'Email adress',
        'creation_date': "Creation date",
        'last_login': "Last login",
        'attributes': "Attributes"
    }

    def arguments(self, parser: ArgumentParser):
        parser.add_argument('username', help="User name to search")

    def run(self, args) -> bool:

        req = UserInfoRequest()
        if not req.is_valid({'username': args.username}):
            for k, v in req.get_errors().items():
                self.tty.write(self.tty.colorizer.red("Error in {}: {}".format(k, v['*'])))
            return False

        record = self.svc_user.get_by_username(args.username)
        if record is None:
            self.tty.write(self.tty.colorizer.red("Error: username not found"))
            return False

        data = record.asdict()
        for field, label in self.fields_labels.items():
            print("{:<16}: {}".format(label, str(data[field])))

        return True


class UserModCmd(UserCommand):
    description = "change user details, including password"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument('username', help="User name of account to change")
        parser.add_argument('-p', '--password', action='store_true', help="Change password")
        parser.add_argument('-a', '--noadmin', action='store_true',
                            help="Remove admin privileges")
        parser.add_argument('-A', '--admin', action='store_true',
                            help="Set admin privileges")
        parser.add_argument('-d', '--disabled', action='store_true',
                            help="Disable user")
        parser.add_argument('-e', '--enabled', action='store_true',
                            help="Enable user")

    def run(self, args) -> bool:

        req = UserInfoRequest()
        if not req.is_valid({'username': args.username}):
            for k, v in req.get_errors().items():
                self.tty.write(self.tty.colorizer.red("Error in {}: {}".format(k, v['*'])))
            return False

        if not args.password and not args.noadmin and not args.admin and not args.enabled and not args.disabled:
            self.tty.write(self.tty.colorizer.red("Error: missing modification options"))
            return False

        if args.noadmin and args.admin:
            self.tty.write(self.tty.colorizer.red("Error: mutually exclusive options for admin privileges"))
            return False

        if args.disabled and args.enabled:
            self.tty.write(self.tty.colorizer.red("Error: mutually exclusive options for account status"))
            return False

        record = self.svc_user.get_by_username(args.username)
        if record is None:
            self.tty.write(self.tty.colorizer.red("Error: username not found"))
            return False

        changes = False
        if args.password:
            not_valid = True
            pwd = ""
            while not_valid:
                pwd = getpass.getpass('New password:')
                if len(pwd) == 0:
                    self.tty.write("operation aborted")
                    return False
                confirmation = getpass.getpass('Confirm password:')

                pwd = pwd.strip()
                confirmation = confirmation.strip()
                if pwd != confirmation:
                    self.tty.write(self.tty.colorizer.red(
                        "password and confirmation don't match, try again or press return to leave"))
                else:
                    not_valid = False

            if len(pwd) > 0:
                changes = True
                if self.svc_auth.update_password(record.username, pwd):
                    # reload record
                    record = self.svc_user.get_by_username(args.username)
                    self.tty.write(self.tty.colorizer.green("> updated user password"))
                else:
                    self.tty.write(self.tty.colorizer.yellow(
                        "> password update failed; maybe the auth backend doesn't support local password changes?"))
                    return False

        if args.noadmin:
            if record.admin:
                changes = True
                self.tty.write(self.tty.colorizer.green("> removed admin privileges from user"))
            record.admin = False

        if args.admin:
            if not record.admin:
                changes = True
                self.tty.write(self.tty.colorizer.green("> added admin privileges to user"))
            record.admin = True

        if args.enabled:
            if not record.active:
                changes = True
                self.tty.write(self.tty.colorizer.green("> enabled user account"))
            record.active = True

        if args.disabled:
            if record.active:
                changes = True
                self.tty.write(self.tty.colorizer.green("> disabled user account"))
            record.active = False

        if changes:
            self.svc_user.update_user(record)
            self.tty.write("User account updated successfully")
        else:
            self.tty.write("User account already conforms to required changes")
        return True

    @property
    def svc_auth(self) -> AuthService:
        return self.get_di().get(DI_SERVICE_MANAGER).get(SVC_AUTH)


class UserListCmd(UserCommand):
    description = "list users"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument('-o', '--offset', type=int, help="Start offset", default=0)
        parser.add_argument('-c', '--count', type=int, help="Records to show", default=0)
        parser.add_argument('-i', '--id', action='store_true', help="Sort by id", default=False)

    def run(self, args) -> bool:

        if args.offset < 0:
            self.tty.write(self.tty.colorizer.red("Error: offset cannot be negative"))
            return False

        if args.count < 0:
            self.tty.write(self.tty.colorizer.red("Error: count cannot be negative"))
            return False

        sort_field = None
        limit = 100
        if not args.id:
            sort_field = UserRecord.username
        if args.count != 0:
            limit = args.count
        details = self.svc_user.list_users(args.offset, limit, sort_field)
        # update limit var for display purposes
        if args.count == 0:
            limit = details[0]
        self.tty.write(self.tty.colorizer.green("Displaying {} of {} users:".format(limit, details[0])))

        for user in details[1]:  # type: UserRecord
            self.tty.write(
                "{:<12}: {}, {}, {} {}".format(user.id, user.username, user.email, user.first_name, user.last_name))

        return True



from rick_db import fieldmapper


@fieldmapper(tablename='acl_role', pk='id_acl_role')
class AclRole:
    id = 'id_acl_role'
    description = 'description'


@fieldmapper(tablename='acl_resource', pk='id_acl_resource')
class AclResource:
    id = 'id_acl_resource'
    description = 'description'


@fieldmapper(tablename='acl_role_resource', pk='id_acl_role_resource')
class AclRoleResource:
    id = 'id_acl_role_resource'
    id_role = 'fk_acl_role'
    id_resource = 'fk_acl_resource'


@fieldmapper(tablename='acl_user_role', pk='id_acl_user_role')
class AclUserRole:
    id = 'id_acl_user_role'
    id_role = 'fk_acl_role'
    id_user = 'fk_user'



from typing import Union

from rick_db import Repository
from rick_db.sql import Sql

from pokie.contrib.auth.dto import UserRecord


class UserRepository(Repository):

    def __init__(self, db):
        super().__init__(db, UserRecord)

    def find_by_username(self, username: str) -> Union[UserRecord, None]:
        key = '__user__:find_by_username'
        sql = self._cache_get(key)
        if not sql:
            sql, _ = self.select() \
                .where(UserRecord.username, '=', username) \
                .assemble()
            self._cache_set(key, sql)
        with self._db.cursor() as c:
            return c.fetchone(sql, [username], cls=UserRecord)

    def list_users(self, offset: int, limit: int, sort_field=None, sort_order=None) -> tuple:
        if not sort_field:
            sort_field = UserRecord.id
        if not sort_order:
            sort_order = Sql.SQL_ASC
        qry = self.select().order(sort_field, sort_order)
        return self.list(qry, limit, offset, UserRecord)



from typing import Any, Optional

from rick.base import Di
from rick.mixin import Injectable

from pokie.contrib.base.repository import ValidatorRepository
from pokie.constants import DI_CONFIG, DI_DB


class ValidatorService(Injectable):

    def __init__(self, di: Di):
        super().__init__(di)
        cfg = di.get(DI_CONFIG)
        self.db = di.get(DI_DB)

        self._cache = None
        if cfg.get('db_cache_metadata', False):
            self._cache = {}

    def id_exists(self, pk_name:str, pk_value, table_name: str, schema: str = None) -> bool:
        """
        Check if pk_value exists on the specified table as primary key value

        :param pk_name: primary key name (optional)
        :param pk_value: primary key value
        :param table_name: table name
        :param schema: optional schema name
        :return: True if id is a valid primary key, False if it doesn't
        """
        if not pk_name:
            pk_name = self.get_pk_field(table_name, schema)
        return self.repo_validator.pk_exists(pk_value, pk_name, table_name, schema)

    def get_pk_field(self, table_name: str, schema: str = None) -> Optional[str]:
        """
        Get primary key field name

        :param table_name:
        :param schema:
        :return: str or None
        """
        key = "{}.{}".format(table_name, schema)
        name = self._cache_get(table_name)
        if name is not None:
            return name

        record = self.db.metadata().table_pk(table_name, schema)
        if record is not None:
            self._cache_add(key, record.field)
            return record.field

        return None

    @property
    def repo_validator(self) -> ValidatorRepository:
        return ValidatorRepository(self.db)

    def _cache_add(self, key, value):
        if self._cache is not None:
            self._cache[key] = value

    def _cache_get(self, key, default=None) -> Any:
        if self._cache is not None:
            if key in self._cache.keys():
                return self._cache[key]
        return default



import uuid

from rick.resource.config import StrOrFile


class BaseConfigTemplate:
    # list of enabled module names
    MODULES = []
    # if true, all endpoints are authenticated by default
    USE_AUTH = True
    # Autentication plugins to use
    AUTH_PLUGINS = ['pokie.contrib.auth.plugin.DbAuthPlugin']

    # Secret key for flask-login hashing
    AUTH_SECRET = uuid.uuid4().hex

    # cache table-related metadata (such as primary key info)
    # development should be false
    DB_CACHE_METADATA = False


class PgConfigTemplate:
    # Postgresql Configuration
    DB_NAME = 'pokie'
    DB_HOST = 'localhost'
    DB_PORT = 5432
    DB_USER = StrOrFile('postgres')
    DB_PASSWORD = StrOrFile('')
    DB_SSL = 'True'


class RedisConfigTemplate:
    # Redis Configuration
    REDIS_HOST = 'localhost'
    REDIS_PORT = 6379
    REDIS_PASSWORD = StrOrFile('')
    REDIS_DB = 0
    REDIS_SSL = '1'



from rick.util.string import snake_to_pascal, snake_to_camel
from rick_db import fieldmapper

from pokie.codegen.spec import TableSpec
from pokie.codegen.textfile import TextBuffer


class RecordGenerator:

    def gen_source(self, spec: TableSpec, camelcase=False, gen: TextBuffer = None, imports=True):
        """
        Generate a RickDB Record class source definition

        :param spec:
        :param camelcase: if True, attributes will be camelCased
        :param gen: optional TextBuffer
        :param imports: include import line
        :return: source code string
        """
        if gen is None:
            gen = TextBuffer()

        if imports is True:
            gen.writeln("from rick_db import fieldmapper", newlines=3)

        gen.writeln("@fieldmapper(tablename='{}', pk='{}', schema='{}')".format(spec.table, spec.pk, spec.schema))
        gen.writeln("class {}Record:".format(snake_to_pascal(spec.table)))
        for field in spec.fields:
            if field.pk is False:
                attr = field.name if camelcase is False else snake_to_camel(field.name)
            else:
                attr = 'id'

            gen.writeln("{name} = '{field}'".format(name=attr, field=field.name), level=1)
        gen.writeln(gen.newline())
        return gen.read()

    def gen_class(self, spec: TableSpec, camelcase=False):
        """
        Generate a RickDB Record class from a spec
        :param spec:
        :param camelcase: if True, attributes will be camelCased
        :return: Record class
        """

        class clsRecord:
            pass

        for field in spec.fields:
            if field.pk is False:
                attr = field.name if camelcase is False else snake_to_camel(field.name)
            else:
                attr = 'id'

            setattr(clsRecord, attr, field.name)

        fieldmapper(clsRecord, spec.pk, tablename=spec.table, schema=spec.schema)
        return clsRecord



import abc
from typing import Optional

from flask_login import UserMixin
from rick.mixin import Injectable


class AuthPluginInterface(Injectable, abc.ABC):
    UPDATE_PASSWORD = 1

    @abc.abstractmethod
    def autenticate(self, username: str, password: str) -> Optional[UserMixin]:
        pass

    @abc.abstractmethod
    def valid_username(self, username: str) -> bool:
        pass

    @abc.abstractmethod
    def update_password(self, username: str, password: str) -> bool:
        pass

    @abc.abstractmethod
    def is_local(self) -> bool:
        pass

    @abc.abstractmethod
    def has_capability(self, capability: int) -> bool:
        pass

    @abc.abstractmethod
    def load_id(self, id_user):
        pass


from typing import List, Optional

from rick_db import Repository
from pokie.contrib.base.dto import SettingsRecord


class SettingsRepository(Repository):

    def __init__(self, db):
        super().__init__(db, SettingsRecord)

    def fetch_by_module(self, module: str) -> List[SettingsRecord]:
        """
        Fetch config entries by module
        :param module: module name
        :return:
        """
        qry = self._cache_get('fetch_by_module')
        if qry is None:
            qry, values = self.select().where(SettingsRecord.module, '=', module).assemble()
            self._cache_set('fetch_by_module', qry)
        else:
            values = [module]
        with self._db.cursor() as c:
            return c.fetchall(qry, values, self._record)

    def fetch_by_key(self, module: str, key: str) -> Optional[SettingsRecord]:
        """
        Fetch settings by key
        :param module: module name
        :param key: key name
        :return:
        """
        qry = self._cache_get('fetch_by_key')
        if qry is None:
            qry, values = self.select() \
                .where(SettingsRecord.module, '=', module) \
                .where(SettingsRecord.key, '=', key) \
                .assemble()
            self._cache_set('fetch_by_key', qry)
        else:
            values = [module, key]
        with self._db.cursor() as c:
            return c.fetchone(qry, values, self._record)

    def upsert(self, module: str, key: str, value: str):
        """
        Insert or update settings entries
        :param module: module name
        :param key: key name
        :param value: value
        :return: id if new record is created
        """
        record = self.fetch_by_key(module, key)
        if record is not None:
            record.value = value
            return self.update(record)
        return self.insert_pk(SettingsRecord(module=module, key=key, value=value))



from pokie.contrib.base.constants import SVC_VALIDATOR, SVC_SETTINGS
from pokie.contrib.base.validators import init_validators
from pokie.core import BaseModule


class Module(BaseModule):
    name = "base"
    description = "Pokie base module"

    cmd = {
        # base commands
        'list': 'pokie.contrib.base.cli.ListCmd',
        'help': 'pokie.contrib.base.cli.HelpCmd',
        'version': 'pokie.contrib.base.cli.VersionCmd',
        'runserver': 'pokie.contrib.base.cli.RunServerCmd',

        # database-related commands
        'dbinit': 'pokie.contrib.base.cli.DbInitCmd',
        'dbcheck': 'pokie.contrib.base.cli.DbCheckCmd',
        'dbupdate': 'pokie.contrib.base.cli.DbUpdateCmd',

        # worker job commands
        'joblist': 'pokie.contrib.base.cli.JobListCmd',
        'jobrun': 'pokie.contrib.base.cli.JobRunCmd',

        # code generation
        'codegen:dto': 'pokie.contrib.base.cli.GenDtoCmd',
        'codegen:request': 'pokie.contrib.base.cli.GenRequestRecordCmd',
    }

    services = {
        # db-related validators
        SVC_VALIDATOR: 'pokie.contrib.base.service.ValidatorService',
        # settings service
        SVC_SETTINGS: 'pokie.contrib.base.service.SettingsService'
    }

    jobs = [
        'pokie.contrib.base.job.IdleJob',
    ]

    def build(self, parent=None):
        init_validators(self.get_di())




from typing import Optional

from flask_login import UserMixin
from rick.base import Di
from rick.mixin import Injectable

from pokie.contrib.auth.constants import SVC_ACL
from pokie.contrib.auth.repository.user import UserRepository
from pokie.constants import DI_SERVICE_MANAGER, DI_DB
from rick.util.datetime import iso8601_now
from pokie.contrib.auth.dto import UserRecord


class AuthUser(UserMixin):
    record = None  # type: UserRecord
    resources = None  # type: List
    roles = None  # type: List
    id = -1

    def __init__(self, usr: UserRecord, _di: Di):
        self.record = usr
        self.id = usr.id
        svc_acl = _di.get(DI_SERVICE_MANAGER).get(SVC_ACL)  # type: AclService
        self.resources = svc_acl.list_user_resource_id(usr.id)
        self.roles = svc_acl.list_user_role_id(usr.id)

    def can_access(self, id_resource: str) -> bool:
        return id_resource in self.resources

    def has_role(self, id_role: int):
        return id_role in self.roles

    @property
    def is_authenticated(self):
        return True

    @property
    def is_active(self):
        return self.record.active

    @property
    def is_anonymous(self):
        return False

    def get_id(self):
        return self.id


class UserService(Injectable):

    def get_by_id(self, id_user: int) -> Optional[UserRecord]:
        """
        Find user by id
        :param id_user:
        :return:
        """
        return self.user_repository.fetch_pk(id_user)

    def get_by_username(self, username: str) -> Optional[UserRecord]:
        """
        Find user by username
        :param username:
        :return:
        """
        return self.user_repository.find_by_username(username)

    def update_lastlogin(self, id_user: int):
        """
        Update user last login timestamp
        :param id_user:
        :return:
        """
        self.user_repository.update(UserRecord(id=id_user, last_login=iso8601_now()))

    def update_password(self, id_user: int, password_hash: str):
        """
        Update user password
        :param id_user:
        :param password_hash:
        :return:
        """
        self.user_repository.update(UserRecord(id=id_user, password=password_hash))

    def add_user(self, record: UserRecord) -> int:
        """
        Creates a new user
        :param record:
        :return:
        """
        return self.user_repository.insert_pk(record)

    def list_users(self, offset, limit, sort_field=None, sort_order=None) -> tuple:
        """
        Returns a tuple with a list of users

        :param offset:
        :param limit:
        :param sort_field:
        :param sort_order:
        :return: (total user count, [records])
        """
        return self.user_repository.list_users(offset, limit, sort_field, sort_order)

    def update_user(self, record: UserRecord):
        return self.user_repository.update(record)

    @property
    def user_repository(self) -> UserRepository:
        return UserRepository(self._di.get(DI_DB))



from typing import List

from rick_db import Repository
from rick_db.sql import Select, Literal, Delete, Insert

from pokie.contrib.auth.dto import AclRole, AclUserRole, AclResource, AclRoleResource


class AclRoleRepository(Repository):

    def __init__(self, db):
        super().__init__(db, AclRole)

    def find_user_roles(self, id_user: int) -> List[AclRole]:
        key = '__acl__:find_user_roles'
        sql = self._cache_get(key)
        if not sql:
            sql, _ = self.select() \
                .join(AclUserRole, AclUserRole.id_role, AclRole, AclRole.id) \
                .where(AclUserRole.id_user, '=', id_user) \
                .order(AclRole.id) \
                .assemble()
            self._cache_set(key, sql)
        with self._db.cursor() as c:
            return c.fetchall(sql, [id_user], cls=AclRole)

    def can_remove(self, id_role: int) -> bool:
        """
        Check if a given role is empty or not in use
        :param id_role:
        :return:
        """
        # check if role references resources
        sql, values = Select(self._dialect) \
            .from_(AclRoleResource, cols={Literal('COUNT(*)'): 'total'}, schema=self._schema) \
            .where(AclRoleResource.id_role, '=', id_role) \
            .assemble()

        with self._db.cursor() as c:
            result = c.fetchone(sql, values)
            if result['total'] != 0:
                return False

        # check if role is referenced by users
        sql, values = Select(self._dialect) \
            .from_(AclUserRole, cols={Literal('COUNT(*)'): 'total'}, schema=self._schema) \
            .where(AclUserRole.id_role, '=', id_role) \
            .assemble()

        with self._db.cursor() as c:
            result = c.fetchone(sql, values)
            if result['total'] != 0:
                return False

        return True

    def truncate(self, id_role: int):
        # delete role resources
        sql, values = Delete(self._dialect) \
            .from_(AclRoleResource) \
            .where(AclRoleResource.id_role, '=', id_role) \
            .assemble()
        self.exec(sql, values)

        # delete user associations
        sql, values = Delete(self._dialect) \
            .from_(AclUserRole) \
            .where(AclUserRole.id_role, '=', id_role) \
            .assemble()
        self.exec(sql, values)

        # finally, delete role
        self.delete_pk(id_role)

    def add_role_resource(self, id_role: int, id_resource: int):
        """
        Associates a resource with a role
        :param id_role:
        :param id_resource:
        :return:
        """
        # check if resource is already in role
        sql, values = Select(self._dialect) \
            .from_(AclRoleResource, schema=self._schema) \
            .where(AclRoleResource.id_role, '=', id_role) \
            .where(AclRoleResource.id_resource, '=', id_resource) \
            .assemble()

        with self._db.cursor() as c:
            result = c.fetchall(sql, values)
            if len(result) > 0:
                return

        sql, values = Insert(self._dialect) \
            .into(AclRoleResource(id_role=id_role, id_resource=id_resource)) \
            .assemble()
        self.exec(sql, values)

    def remove_role_resource(self, id_role: int, id_resource: int):
        """
        Removes an association between a resource and a role
        :param id_role:
        :param id_resource:
        :return:
        """
        # check if resource is already in role
        sql, values = Select(self._dialect) \
            .from_(AclRoleResource, schema=self._schema) \
            .where(AclRoleResource.id_role, '=', id_role) \
            .where(AclRoleResource.id_resource, '=', id_resource) \
            .assemble()

        with self._db.cursor() as c:
            result = c.fetchall(sql, values)
            if len(result) == 0:
                return

        # delete resource from role association
        sql, values = Delete(self._dialect) \
            .from_(AclRoleResource) \
            .where(AclRoleResource.id_role, '=', id_role) \
            .where(AclRoleResource.id_resource, '=', id_resource) \
            .assemble()
        self.exec(sql, values)

    def remove_user_role(self, id_user: int, id_role: int):
        """
        Removes an association between a user and a role
        :param id_user:
        :param id_role:
        :return:
        """
        # check if role is associated with user
        sql, values = Select(self._dialect) \
            .from_(AclUserRole, schema=self._schema) \
            .where(AclUserRole.id_role, '=', id_role) \
            .where(AclUserRole.id_user, '=', id_user) \
            .assemble()

        with self._db.cursor() as c:
            result = c.fetchall(sql, values)
            if len(result) == 0:
                return

        # delete role from user
        sql, values = Delete(self._dialect) \
            .from_(AclUserRole) \
            .where(AclUserRole.id_role, '=', id_role) \
            .where(AclUserRole.id_user, '=', id_user) \
            .assemble()
        self.exec(sql, values)

    def add_user_role(self, id_user, id_role):
        """
        Associates a user with a role
        :param id_user:
        :param id_role:
        :return:
        """
        # check if role is associated with user
        sql, values = Select(self._dialect) \
            .from_(AclUserRole, schema=self._schema) \
            .where(AclUserRole.id_role, '=', id_role) \
            .where(AclUserRole.id_user, '=', id_user) \
            .assemble()

        with self._db.cursor() as c:
            result = c.fetchall(sql, values)
            if len(result) > 0:
                return

        # insert association record
        sql, values = Insert(self._dialect).into(AclUserRole(id_role=id_role, id_user=id_user)).assemble()
        self.exec(sql, values)


class AclResourceRepository(Repository):

    def __init__(self, db):
        super().__init__(db, AclResource)

    def find_user_resources(self, id_user: int) -> List[AclResource]:
        key = '__acl__:find_user_resources'
        sql = self._cache_get(key)
        if not sql:
            sql, _ = self.select() \
                .join(AclRoleResource, AclRoleResource.id_resource, AclResource, AclResource.id) \
                .join(AclUserRole, AclUserRole.id_role, AclRoleResource, AclRoleResource.id_role) \
                .where(AclUserRole.id_user, '=', id_user) \
                .assemble()
            self._cache_set(key, sql)

        with self._db.cursor() as c:
            return c.fetchall(sql, [id_user], cls=AclRole)

    def find_by_role(self, id_role: int) -> List[AclResource]:
        """
        List resources for a given role
        :param id_role:
        :return:
        """
        key = '__acl__:find_by_role'
        sql = self._cache_get(key)
        if not sql:
            sql, _ = self.select() \
                .join(AclRoleResource, AclRoleResource.id_resource, AclResource, AclResource.id) \
                .where(AclRoleResource.id_role, '=', id_role) \
                .order(AclResource.id) \
                .assemble()
            self._cache_set(key, sql)

        with self._db.cursor() as c:
            return c.fetchall(sql, [id_role], cls=AclResource)

    def can_remove(self, id_resource: int):
        """
        Check if a given resource can be removed
        :param id_resource:
        :return:
        """
        # check if role references resources
        sql, values = Select(self._dialect) \
            .from_(AclRoleResource, cols={Literal('COUNT(*)'): 'total'}, schema=self._schema) \
            .where(AclRoleResource.id_resource, '=', id_resource) \
            .assemble()

        with self._db.cursor() as c:
            result = c.fetchone(sql, values)
            return result['total'] == 0

    def truncate(self, id_resource: int):
        # delete role resources
        sql, values = Delete(self._dialect) \
            .from_(AclRoleResource) \
            .where(AclRoleResource.id_resource, '=', id_resource) \
            .assemble()
        self.exec(sql, values)

        # finally, delete resource
        self.delete_pk(id_resource)



from argparse import ArgumentParser
from typing import List

from rick.mixin import Injectable, Runnable
from rick.util.loader import load_class

from pokie.constants import DI_APP
from pokie.contrib.base.cli.base import BaseCommand


class JobBaseCmd(BaseCommand):

    def get_jobs(self) -> dict:
        result = {}
        for module_name, module in self.get_di().get(DI_APP).modules.items():
            jobs = getattr(module, 'jobs', [])
            if len(jobs) > 0:
                result[module_name] = jobs
        return result


class JobListCmd(JobBaseCmd):
    description = "list registered job workers"

    def run(self, args) -> bool:
        for name, jobs in self.get_jobs().items():
            self.tty.write("Worker Jobs for module {}:".format(name))
            for job in jobs:
                self.tty.write(self.tty.colorizer.white("   {}".format(job), attr='bold'))

        return True

class JobRunCmd(JobBaseCmd):
    description = "run  all job workers"

    def run(self, args) -> bool:
        joblist = []
        di = self.get_di()

        # prepare job list
        for module_name, jobs in self.get_jobs().items():
            for job_name in jobs:
                self.tty.write("Preparing job  '{}'...".format(job_name))
                job = load_class(job_name)
                if job is None:
                    raise ValueError("Non-existing job class '{}' in module {}".format(job_name, module_name))
                if not issubclass(job, (Injectable, Runnable)):
                    raise RuntimeError("Class '{}' must implement Injectable, Runnable interfaces")
                joblist.append(job(di))

        # run job list
        self.tty.write("\nRunning jobs, press CTRL+C to abort...")
        while True:
            for job in joblist:
                job.run(di)

        # unreachable
        return True


# Version
POKIE_VERSION = ["0", "1", "2"]


def get_version():
    return ".".join(POKIE_VERSION)

# Http Codes
HTTP_OK = 200
HTTP_BADREQ = 400
HTTP_NOAUTH = 401
HTTP_FORBIDDEN = 403
HTTP_INTERNAL_ERROR = 500

# DI Keys
DI_CONFIG = 'config'
DI_FLASK = 'app'
DI_APP = 'main'
DI_MODULES = 'modules'
DI_SERVICE_MANAGER = 'svc_manager'
DI_DB = 'db'
DI_REDIS = 'redis'
DI_EVENT = 'event_manager'
DI_TTY = 'tty'

# DB Configuration
CFG_DB_NAME = 'db_name'
CFG_DB_HOST = 'db_host'
CFG_DB_PORT = 'db_port'
CFG_DB_USER = 'db_user'
CFG_DB_PASSWORD = 'db_password'
CFG_DB_SSL = 'db_ssl'

# Redis Configuration
CFG_REDIS_HOST = 'redis_host'
CFG_REDIS_PORT = 'redis_port'
CFG_REDIS_PASSWORD = 'redis_password'
CFG_REDIS_DB = 'redis_db'
CFG_REDIS_SSL = 'redis_ssl'

# Auth Configuration
CFG_AUTH_SECRET = 'auth_secret'


from time import sleep

from rick.base import Di
from rick.mixin import Injectable, Runnable

from pokie.constants import DI_CONFIG


class IdleJob(Injectable, Runnable):
    DEFAULT_IDLE_INTERVAL = 15  # 15s between runs

    def __init__(self, di: Di):
        super().__init__(di)
        cfg = di.get(DI_CONFIG)
        self.interval = int(cfg.get('job_idle_interval', self.DEFAULT_IDLE_INTERVAL))

    def run(self, di: Di):
        sleep(self.interval)



from rick.form import RequestRecord
from flask import request
from .helpers import ParseListError, parse_list_parameters
from pokie.rest import RestService
from pokie.constants import DI_SERVICE_MANAGER
from inspect import isclass


class RestMixin:
    request_class = None  # type: rick.form.RequestRecord
    record_class = None
    search_fields = None  # type: List
    svc_name = None

    def get(self, id_record=None):
        """
        Read single record by id
        :param id_record:
        :return:
        """
        if id_record is None:
            return self.list()

        record = self.svc.get(id_record)
        if record is None:
            return self.not_found()

        return self.success(record)

    def list(self):
        """
        Query records
        :return:
        """
        search_fields = self.search_fields if self.search_fields is not None else []
        text, match, limit, offset, sort = parse_list_parameters(request.args, self.record_class)
        try:
            count, data = self.svc.list(
                search_fields=search_fields,
                search_text=text,
                match_fields=match,
                limit=limit,
                offset=offset,
                sort_fields=sort)
            result = {
                'total': count,
                'rows': data
            }
            return self.success(result)
        except ParseListError as e:
            return self.error(str(e))

    def post(self):
        """
        Create Record
        :return:
        """
        data = request.json
        if data is None or len(data) == 0:
            return self.empty_body()

        req = self.request_factory
        if not req.is_valid(request.json):
            return self.request_error(req)

        record = req.bind(self.record_class)
        self.svc.insert(record)
        return self.success()

    def put(self, id_record):
        """
        Update Record
        :return:
        """
        data = request.json
        if data is None or len(data) == 0:
            return self.empty_body()

        if not self.svc.exists(id_record):
            return self.not_found()

        req = self.request_factory
        if not req.is_valid(request.json):
            return self.request_error(req)

        record = req.bind(self.record_class)
        self.svc.update(id_record, record)
        return self.success()

    def delete(self, id_record):
        """
        delete Record by id
        :param id_record:
        :return:
        """
        if not self.svc.exists(id_record):
            return self.not_found()

        self.svc.delete(id_record)
        return self.success()

    @property
    def request_factory(self) -> RequestRecord:
        if isclass(self.request_class):
            return self.request_class()
        return self.request_class

    @property
    def svc(self) -> RestService:
        mgr = self.di.get(DI_SERVICE_MANAGER)
        if self.svc_name is None:
            svc_name = "svc.rest.{}.{}".format(self.__module__, str(self.record_class.__name__).replace('Record', '', 1))
            if mgr.contains(svc_name):
                return mgr.get(svc_name)

            # register new service that relies on a RestService instance
            mgr.add(svc_name, 'pokie.rest.RestService')
            # get new service to patch it
            svc = mgr.get(svc_name)
            # patch it
            svc.set_record_class(self.record_class)
            return svc

        svc = mgr.get(self.svc_name)
        if not isinstance(svc, RestService):
            raise RuntimeError("Service '{}' does not implement RestService mixin")
        return svc



import os
import sys
from argparse import ArgumentParser
from typing import List
from collections import OrderedDict

import DistUpgrade.apt_btrfs_snapshot
from flask import Flask
from rick.base import Di, Container, MapLoader
from rick.event import EventManager
from rick.util.loader import load_class
from rick.resource.console import ConsoleWriter

from pokie.constants import DI_CONFIG, DI_SERVICE_MANAGER, DI_FLASK, DI_APP, DI_EVENT, DI_TTY
from .module import BaseModule
from .command import CliCommand
from ..util.cli_args import ArgParser


class FlaskApplication:
    module_file_name = 'module'  # module class file name
    module_class_name = 'Module'  # default module class name

    system_modules = ['pokie.contrib.base', ]  # system modules to always be included

    def __init__(self, cfg: Container):
        self.di = Di()
        self.app = None
        self.modules = {}  # app module list

        self.di.add(DI_CONFIG, cfg)
        self.di.add(DI_APP, self)
        self.cfg = cfg

    def build(self, module_list: list, factories: List = None) -> Flask:
        """
        Build the application

        Factories is a list of optional callables to assemble functionality on top of Di, eg. database connection,
        cache, logging, etc. Factories are called *before* modules are initialized, to ensure all required dependencies
        are available

        :param module_list: list of module names to initialize
        :param factories: optional list of callables to be initialized with the application
        :return:
        """
        if not factories:
            factories = []

        self.app = Flask(type(self).__name__)
        self.app.di = self.di
        self.di.add(DI_FLASK, self.app)

        # load modules
        self.modules = {}
        module_list = [*self.system_modules, *module_list]
        for name in module_list:
            cls = load_class("{}.{}.{}".format(name, self.module_file_name, self.module_class_name))
            if cls is None:
                raise RuntimeError("build(): cannot load module '{}' - Module() class not found".format(name))
            if not issubclass(cls, BaseModule):
                raise RuntimeError("build(): Class Module on '{}' must extend BaseModule".format(name))
            if name in self.modules.keys():
                raise ValueError("build(): Module named '{}' already exists".format(name))
            self.modules[name] = cls(self.di)

        # build service map
        svc_map = {}
        for name, m in self.modules.items():
            services = getattr(m, 'services', {})
            if type(services) is dict:
                svc_map.update(services)
            else:
                raise RuntimeError(
                    "build(): cannot load service map from module '{}'; attribute must be of type dict".format(name))
        # register service mapper
        self.di.add(DI_SERVICE_MANAGER, MapLoader(self.di, svc_map))

        # run factories
        for factory in factories:
            if type(factory) is str:
                # if factory is string, assume it is a path to a callable
                factory = load_class(factory)
            if not callable(factory):
                raise RuntimeError("build(): non-callable or non-existing factory")
            else:
                factory(self.di)

        # parse events from modules
        evt_mgr = EventManager()
        for _, module in self.modules.items():
            module_events = getattr(module, 'events', None)
            if isinstance(module_events, dict):
                for evt_name, evt_details in module_events.items():
                    for priority, handlers in evt_details.items():
                        for handler in handlers:
                            evt_mgr.add_handler(evt_name, handler, int(priority))

        self.di.add(DI_EVENT, evt_mgr)

        # initialize TTY
        self.di.add(DI_TTY, ConsoleWriter())

        # initialize modules
        for _, module in self.modules.items():
            module.build(self)

        return self.app

    def http(self, **kwargs):
        self.app.run(**kwargs)

    def cli(self, **kwargs):
        """
        Execute CLI commands
        :param kwargs: optional parameters for ArgumentParse
        :return:
        """
        if 'writer' in kwargs.keys():
            tty = kwargs['writer']
        else:
            tty = ConsoleWriter()

        # default command when no args detected
        command = 'list'
        # extract command if specified
        if len(sys.argv) > 1:
            command = str(sys.argv[1])

        if 'add_help' not in kwargs.keys():
            kwargs['add_help'] = False
        if 'usage' not in kwargs.keys():
            kwargs['usage'] = "{} {} [OPTIONS...]".format(os.path.basename(sys.argv[0]), command)

        parser = ArgParser(**kwargs)

        # lookup handler
        for _, module in self.modules.items():
            if command in module.cmd.keys():
                handler = load_class(module.cmd[command])
                if not handler:
                    raise RuntimeError("cli(): handler class '{}' not found".format(module.cmd[command]))
                if not issubclass(handler, CliCommand):
                    raise RuntimeError("cli(): command handler does not extend CliCommand")
                handler = handler(self.di, writer=tty)  # type: CliCommand
                handler.arguments(parser)
                args = parser.parse_args(sys.argv[2:])
                if parser.failed:
                    # invalid/insufficient args
                    tty.error(parser.error_message)
                    parser.print_help(tty.stderr)
                    exit(1)

                if not handler.run(args):
                    exit(1)
                exit(0)

        # command not found
        tty.error("error executing '{}': command not found".format(command))
        exit(2)



from rick_db import fieldmapper


@fieldmapper(tablename='user', pk='id_user')
class UserRecord:
    id = 'id_user'
    active = 'active'
    admin = 'admin'
    username = 'username'
    first_name = 'first_name'
    last_name = 'last_name'
    email = 'email'
    creation_date = 'creation_date'
    last_login = 'last_login'
    password = 'password'
    attributes = 'attributes'



from pokie.contrib.auth.constants import SVC_USER, SVC_ACL, SVC_AUTH
from pokie.core import BaseModule


class Module(BaseModule):
    name = "auth"
    description = "Authentication module"

    services = {
        SVC_AUTH: 'pokie.contrib.auth.service.AuthService',
        SVC_ACL: 'pokie.contrib.auth.service.AclService',
        SVC_USER: 'pokie.contrib.auth.service.UserService',
    }

    cmd = {
        # user-related operations
        'usercreate': 'pokie.contrib.auth.cli.UserCreateCmd',
        'userinfo': 'pokie.contrib.auth.cli.UserInfoCmd',
        'usermod': 'pokie.contrib.auth.cli.UserModCmd',
        'userlist': 'pokie.contrib.auth.cli.UserListCmd',
        'acluserrole': 'pokie.contrib.auth.cli.AclUserRoleCmd',

        # acl operations
        'aclrolelist': 'pokie.contrib.auth.cli.AclRoleListCmd',
        'aclrolecreate': 'pokie.contrib.auth.cli.AclRoleCreateCmd',
        'aclroleremove': 'pokie.contrib.auth.cli.AclRoleRemoveCmd',
        'aclroleinfo': 'pokie.contrib.auth.cli.AclRoleInfoCmd',
        'aclrolelink': 'pokie.contrib.auth.cli.AclRoleLinkCmd',
        'aclroleunlink': 'pokie.contrib.auth.cli.AclRoleUnlinkCmd',

        'aclresourcelist': 'pokie.contrib.auth.cli.AclResourceListCmd',
        'aclresourcecreate': 'pokie.contrib.auth.cli.AclResourceCreateCmd',
        'aclresourcelink': 'pokie.contrib.auth.cli.AclResourceLinkCmd',
        'aclresourceunlink': 'pokie.contrib.auth.cli.AclResourceUnlinkCmd',
    }


    def build(self, parent=None):
        pass



import inspect
import os
from argparse import ArgumentParser
from pathlib import Path
from typing import Optional, List

from rick_db.conn import Connection
from rick_db.sql.dialect import PgSqlDialect
from rick_db.util.pg import PgInfo

from pokie.codegen.pg import PgTableSpec, RecordGenerator, RequestGenerator
from pokie.constants import DI_DB, DI_APP
from pokie.core import CliCommand


class CodeGenCommand(CliCommand):

    def get_db(self) -> Optional[Connection]:
        result = None
        di = self.get_di()
        if di.has(DI_DB):
            result = di.get(DI_DB)
        return result

    def is_supported(self, db) -> bool:
        if isinstance(db, Connection):
            return isinstance(db.dialect(), PgSqlDialect)
        return False

    def parse_table_list(self, mgr: PgInfo, table_expr) -> [List, str]:
        schema = mgr.SCHEMA_DEFAULT
        table_list = []
        table_expr = table_expr.split('.', 1)
        table = table_expr[0]
        if len(table_expr) > 1:
            schema = table_expr[0]
            table = table_expr[1]

        all_tables = []
        for tbl in mgr.list_database_tables(schema):
            all_tables.append(tbl.name)

        if table == '*':
            table_list = all_tables
        else:
            if table not in all_tables:
                self.tty.error("table with name '{}' not found in schema '{}'".format(table, schema))
                return []
            table_list.append(table)

        return table_list, schema

    def pg_gen_dto(self, db, table_expr, dest_file, camel_case=False) -> bool:

        pg = PgTableSpec(db)
        table_list, schema = self.parse_table_list(pg.manager(), table_expr)
        if len(table_list) == 0:
            return False

        gen = RecordGenerator()
        first = True
        result = []
        for name in table_list:
            self.tty.write(self.tty.colorizer.white("generating dto for {}.{}...".format(schema, name), attr='bold'))
            spec = pg.table_spec(name, schema)
            result.append(gen.gen_source(spec, camelcase=camel_case, imports=first))
            first = False

        if dest_file is None:
            self.tty.write("\n".join(result))
        else:
            self.tty.write(self.tty.colorizer.white("writing to file '{}'...".format(dest_file)))
            with open(dest_file, 'w') as f:
                f.write("\n".join(result))

            self.tty.write(self.tty.colorizer.green("success!"))
        return True

    def pg_gen_request(self, db, table_expr, dest_file, camelcase_id=False, camelcase_cols=False) -> bool:

        pg = PgTableSpec(db)
        table_list, schema = self.parse_table_list(pg.manager(), table_expr)
        if len(table_list) == 0:
            return False

        gen = RequestGenerator()
        first = True
        result = []
        for name in table_list:
            self.tty.write(self.tty.colorizer.white("generating RequestRecord class for {}.{}...".format(schema, name),
                                                    attr='bold'))
            spec = pg.table_spec(name, schema)
            result.append(gen.gen_source(spec, camelcase=camelcase_id,db_camelcase=camelcase_cols, imports=first))
            first = False

        if dest_file is None:
            self.tty.write("\n".join(result))
        else:
            self.tty.write(self.tty.colorizer.white("writing to file '{}'...".format(dest_file)))
            with open(dest_file, 'w') as f:
                f.write("\n".join(result))

            self.tty.write(self.tty.colorizer.green("success!"))
        return True


class GenDtoCmd(CodeGenCommand):
    description = "generate dto class from database table"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument('table', type=str, help="source table or view for dto generation")
        parser.add_argument('-f', dest='file', type=str, help="destination file")
        parser.add_argument('-c', '--camelcase', action="store_true", default=False, help="camelCase attributes")

    def run(self, args) -> bool:
        db = self.get_db()
        if not self.is_supported(db):
            self.tty.error("database code generation is only supported with PostgreSQL")
            return False

        if args.file is not None:
            dest_file = Path(args.file)
            if dest_file.exists():
                self.tty.error("destination file already exists")
                return False

        return self.pg_gen_dto(db, args.table, args.file, args.camelcase)


class GenRequestRecordCmd(CodeGenCommand):
    description = "generate RequestRecord class from database table"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument('table', type=str, help="source table or view for request record generation")
        parser.add_argument('-f', dest='file', type=str, help="destination file")
        parser.add_argument('-c', '--camelcase-names', action="store_true", default=False, help="camelCase field names")
        parser.add_argument('-C', '--camelcase-cols', action="store_true", default=False, help="camelCase bind columns")

    def run(self, args) -> bool:
        db = self.get_db()
        if not self.is_supported(db):
            self.tty.error("database code generation is only supported with PostgreSQL")
            return False

        if args.file is not None:
            dest_file = Path(args.file)
            if dest_file.exists():
                self.tty.error("destination file already exists")
                return False

        return self.pg_gen_request(db, args.table, args.file, args.camelcase_names, args.camelcase_cols)



import uuid

from flask_login import LoginManager
from rick.base import Di

from pokie.constants import DI_CONFIG, DI_SERVICE_MANAGER, CFG_AUTH_SECRET, DI_FLASK
from pokie.contrib.auth.constants import SVC_AUTH


def FlaskLogin(_di: Di):
    """
    FlaskLogin factory
    :param _di:
    :return:
    """
    cfg = _di.get(DI_CONFIG)
    app = _di.get(DI_FLASK)
    app.secret_key = cfg.get(CFG_AUTH_SECRET, uuid.uuid4().hex)
    login_manager = LoginManager()
    login_manager.init_app(app)

    @login_manager.user_loader
    def load_user(user_id):
        # restores user profile from auth service
        return _di.get(DI_SERVICE_MANAGER).get(SVC_AUTH).load_id(user_id)



import abc

from rick.mixin import Injectable


class BaseModule(Injectable, abc.ABC):
    # base module name; must be unique
    name = ""
    # module extended description
    description = ""
    # service mapper
    services = {}
    # cli command mapper
    cmd = {}

    @abc.abstractmethod
    def build(self, parent=None):
        """
        Initialize module internals
        :param parent: FlaskApplication instance
        :return:
        """
        pass



from rick.base import Di
from rick.resource.redis import RedisCache

from pokie.constants import CFG_REDIS_HOST, DI_CONFIG, CFG_REDIS_PORT, DI_REDIS, CFG_REDIS_PASSWORD, CFG_REDIS_DB, \
    CFG_REDIS_SSL


def RedisFactory(_di: Di):
    """
    REDIS factory
    Note: The connection is only created when the resource is accessed on Di
    :param _di:
    :return:
    """

    @_di.register(DI_REDIS)
    def _factory(_di: Di):
        cfg = _di.get(DI_CONFIG)
        redis_cfg = {
            'host': cfg.get(CFG_REDIS_HOST, 'localhost'),
            'port': int(cfg.get(CFG_REDIS_PORT, 6379)),
            'password': cfg.get(CFG_REDIS_PASSWORD, ''),
            'db': int(cfg.get(CFG_REDIS_DB, 0)),
            'ssl': True if cfg.get(CFG_REDIS_SSL, None) else False,
        }
        return RedisCache(**redis_cfg)



from .user import UserCreateCmd, UserInfoCmd, UserModCmd, UserListCmd
from .acl import AclRoleListCmd, AclRoleCreateCmd, AclRoleRemoveCmd, AclRoleInfoCmd, AclRoleLinkCmd, AclRoleUnlinkCmd
from .acl import AclResourceListCmd, AclResourceCreateCmd, AclResourceLinkCmd, AclResourceUnlinkCmd, AclUserRoleCmd




from rick.base import Di
from rick_db.conn.pg import PgConnection

from pokie.constants import CFG_DB_NAME, CFG_DB_HOST, CFG_DB_PORT, CFG_DB_USER, CFG_DB_PASSWORD, CFG_DB_SSL, DI_DB, \
    DI_CONFIG


def PgSqlFactory(_di: Di):
    """
    PostgreSQL connection factory
    Note: The connection is only created when the resource is accessed on Di
    """

    @_di.register(DI_DB)
    def _factory(_di: Di):
        cfg = _di.get(DI_CONFIG)
        db_cfg = {
            'dbname': cfg.get(CFG_DB_NAME, 'postgres'),
            'host': cfg.get(CFG_DB_HOST, 'localhost'),
            'port': int(cfg.get(CFG_DB_PORT, 5432)),
            'user': cfg.get(CFG_DB_USER, 'postgres'),
            'password': cfg.get(CFG_DB_PASSWORD, ''),
            'sslmode': None if not cfg.get(CFG_DB_SSL, '1') else 'require'
        }
        return PgConnection(**db_cfg)
